#
#
# Auxiliary functions that:
#  - fit the model
#  - calculate error of single observation
#  - calculate error of the model
#  - predicts new observations
#  

error.fun.binomial <- function(val, ydata, threshold) {
  return(abs(val - ydata) < threshold) 
}

predict.fun.binomial <- function(object, newx) {
  predict(object = object, newx = newx, s = 'lambda.min', type = 'response')
}

model.error.fun.binomial <- function(ydata.predicted, ydata) {
  mean(sqrt((ydata - ydata.predicted)^2))
}

fit.model.fun.binomial <- function(xdata, ydata) {
  suppressWarnings(cv.glmnet(xdata, ydata, alpha = 0, family = 'binomial', lambda = c(0,1e-7, 1e-5, 1e-3, 1e-1), mc.cores = 2))
}

# 
#' RANSAC
#'
#' @param xdata co-variates (rows: observations | coluns: co-variates)
#' @param ydata response
#' @param n minimum number of observations that will be fitted in the model
#' @param threshold error threshold to determine if observation is inliner
#' @param good.fit.perct discard models that have less than this percentage of inliners
#' @param k number of times model is run
#' @param error.fun 
#' @param predict.fun 
#' @param model.error.fun 
#' @param fit.model.fun 
#' @param mc.cores 
#'
#' @return
#' @export
#'
#' @examples
ransac <- function(xdata, ydata, n, threshold, good.fit.perct, 
                   k = 100, 
                   error.fun       = error.fun.binomial, 
                   predict.fun     = predict.fun.binomial,
                   model.error.fun = model.error.fun.binomial,
                   fit.model.fun   = fit.model.fun.binomial,
                   mc.cores = 1) {
  best.error <- Inf
  best.model <- NA
  flog.debug('Starting ransac...')
  #
  # Using parallel computing to speed up
  error.array <- mclapply(seq(k), function(ix) {
    flog.debug('Running iteration %d', ix)
    #
    # draw a random sample from the dataset with size n
    xdata.min.ix <- sample(seq(nrow(xdata)), n)
    xdata.min <- xdata[xdata.min.ix, ]
    ydata.min <- ydata[xdata.min.ix]
    #
    # fit the model using the random sample
    my.fit <- fit.model.fun(xdata.min, ydata.min)
    flog.debug('  finished the fitting %d', ix)
    #
    # test all the observations outside the random dataset
    #  if they are inliners of the model
    test.inliners.ix <- seq(nrow(xdata))[-xdata.min.ix]
    # TODO: drop the sapply and do this with matrices
    res <- sapply(test.inliners.ix, function(test.ix) {
      my.prediction <- predict.fun(object = my.fit, newx = t(xdata[test.ix,]))
      error.fun(my.prediction, ydata[test.ix], threshold)
    })
    also.inliners.ix <- test.inliners.ix[res]
    flog.debug('  also inliners: %d + %d = %d / %d', 
               length(also.inliners.ix), 
               nrow(xdata.min), 
               length(also.inliners.ix) + nrow(xdata.min),
               nrow(xdata))
    #
    # Check if the model is good
    if (length(also.inliners.ix) >= nrow(xdata) * good.fit.perct) {
      # this implies that we may have found a good model
      #  now test how good it is
      xdata.ix     <- c(also.inliners.ix, xdata.min.ix)
      my.residuals <- predict.fun(object = my.fit, newx = xdata[xdata.ix,])
      this.error   <- model.error.fun(ydata[xdata.ix], my.residuals)
      flog.debug('  good fit!! error = %f', this.error)
      return(list(this.error = this.error, this.model = my.fit))
    }
    flog.debug('  did not have a good fit')
    return(list())
  }
  # parameters for mclapply
  , mc.cores = mc.cores, mc.allow.recursive = F, mc.cleanup = T,
  mc.set.seed = T)
  flog.debug('  Finished running all iterations, consolidating...')
  #
  # Consolidate results finding only the best one
  lapply(seq_along(error.array), function(ix) {
    el <- error.array[[ix]]
    if (!is.null(el) && !is.na(el) && length(el) > 1 && 
        !is.null(el$this.error) && el$this.error < best.error) {
      best.error <<- el$this.error
      best.model <<- el$this.model
    }
  })
  return(list(best.model = best.model, best.error = best.error, errors = error.array ))
}

#
#
#
print.ransac.results <- function(results.ransac, xdata, ydata, name = '') {
  if (results.ransac$best.error == Inf) {
    flog.info('Could not find a good fit')
  } else {
    models <- list()
    models$ransac <- results.ransac$best.model
    models$glmnet <- cv.glmnet(xdata, ydata, family = 'binomial')
    #
    my.fits <- list()
    my.fits$ransac <- predict(models$ransac, newx = xdata, s = 'lambda.min', type = 'response')
    my.fits$glmnet <- predict(models$glmnet, newx = xdata, s = 'lambda.min', type = 'response', lambda = c(0, 0.000000000000000001))
    #
    my.rmse <- list()
    my.rmse$ransac <- model.error.fun.binomial(my.fits$ransac, ydata)
    my.rmse$glmnet <- model.error.fun.binomial(my.fits$glmnet, ydata)
    #
    my.df <- data.frame()
    my.df <- rbind(my.df, data.frame(ix = seq(nrow(xdata)), val = (ydata - my.fits$ransac), type = rep('ransac', nrow(xdata))))
    my.df <- rbind(my.df, data.frame(ix = seq(nrow(xdata)), val = (ydata - my.fits$glmnet), type = rep('glmnet', nrow(xdata))))
    my.df$type <- factor(my.df$type)
    colnames(my.df) <- c('ix', 'value', 'type')
    #
    my.df.class <- data.frame()
    ix.sorted <- sort(ydata, index.return = T)$ix
    my.df.class <- rbind(my.df.class, data.frame(ix = seq(nrow(xdata)), val = (my.fits$ransac[ix.sorted]), type = rep('ransac', nrow(xdata))))
    my.df.class <- rbind(my.df.class, data.frame(ix = seq(nrow(xdata)), val = (my.fits$glmnet[ix.sorted]), type = rep('glmnet', nrow(xdata))))
    my.df.class <- rbind(my.df.class, data.frame(ix = seq(nrow(xdata)), val = (ydata[ix.sorted]),          type = rep('ydata',  nrow(xdata))))
    my.df.class$type <- factor(my.df.class$type)
    colnames(my.df.class) <- c('ix', 'value', 'type')
    #
    # misclassifications
    miscl <- list()
    miscl$ransac <- sum(abs(my.fits$ransac - ydata) > .5)
    miscl$glmnet <- sum(abs(my.fits$glmnet - ydata) > .5)
    title.str <- paste0(name, ' : ransac misclass. = ', miscl$ransac, ' : glmnet misclass. = ', miscl$glmnet)
    #
    g <- ggplot(data = my.df.class) + theme_minimal() + ggtitle(title.str) + ylab('Classification') + xlab('Model')
    g <- g + geom_point(aes(ix, value, color = type))
    print(g)
    #
    g <- ggplot(data = my.df) + theme_minimal() + ggtitle(title.str) + ylab('Error') + xlab('Model')
    g <- g + geom_quasirandom(aes(type, value, color = type))
    print(g)
    #
    flog.info('RMSE:\n   ransac = %f\n   glmnet = %f', my.rmse$ransac, my.rmse$glmnet)
  }
}